@@ -46,17 +46,19 @@
 #ifdef ENABLE_ACCESSIBILITY
 #include "layout/htmlboxtable.h"
 #include "a11y/htmlviewaccessiblefactory.h"
-#include "a11y/htmlboxaccessiblefactory.h"
-#include "a11y/htmlboxblockaccessiblefactory.h"
-#include "a11y/htmlboxembeddedaccessiblefactory.h"
-#include "a11y/htmlboxtableaccessiblefactory.h"
-#include "a11y/htmlboxtextaccessiblefactory.h"
+#include "a11y/htmlaccessiblefactory.h"
+#include "a11y/htmlboxblockaccessible.h"
+#include "a11y/htmlboxembeddedaccessible.h"
+#include "a11y/htmlboxaccessible.h"
+#include "a11y/htmlboxtableaccessible.h"
+#include "a11y/htmlboxtextaccessible.h"
 #endif
 
 enum {
 	MOVE_CURSOR,
 	REQUEST_OBJECT,
 	ON_URL,
+	ACTIVATE,
 	LAST_SIGNAL
 };
 
@@ -68,9 +70,18 @@ static GtkLayoutClass *parent_class = NULL;
 
 static gboolean html_view_remove_layout_box (HtmlView *view, DomNode *node);
 static void html_view_try_jump (HtmlView *view);
+static void html_view_set_saved_focus (HtmlView *view);
+static DomElement* html_view_get_saved_focus (HtmlView *view);
+static void html_view_focus_element (HtmlView *view);
 
 #ifdef ENABLE_ACCESSIBILITY
 static AtkObject *  html_view_get_accessible (GtkWidget *widget);
+
+HTML_ACCESSIBLE_FACTORY (HTML_TYPE_BOX_BLOCK_ACCESSIBLE, html_box_block_accessible)
+HTML_ACCESSIBLE_FACTORY (HTML_TYPE_BOX_EMBEDDED_ACCESSIBLE, html_box_embedded_accessible)
+HTML_ACCESSIBLE_FACTORY (HTML_TYPE_BOX_ACCESSIBLE, html_box_accessible)
+HTML_ACCESSIBLE_FACTORY (HTML_TYPE_BOX_TABLE_ACCESSIBLE, html_box_table_accessible)
+HTML_ACCESSIBLE_FACTORY (HTML_TYPE_BOX_TEXT_ACCESSIBLE, html_box_text_accessible)
 #endif
 
 typedef enum {
@@ -214,6 +225,10 @@ html_view_relayout (HtmlView *view)
 		gtk_idle_remove (view->relayout_idle_id);
 		view->relayout_idle_id = 0;
 	}
+	if (GTK_WIDGET_HAS_FOCUS (view) && 
+	    view->document->focus_element == NULL) {
+		html_view_focus (HTML_VIEW (view), GTK_DIR_TAB_FORWARD);
+	}
 }
 
 static gint
@@ -345,9 +360,12 @@ html_view_insert_node (HtmlView *view, DomNode *node)
 
 		html_box_handle_html_properties (new_box, node->xmlnode);
 		
-		if (parent_box == NULL)
+		if (parent_box == NULL) {
+			html_view_layout_tree_free (view, view->root);
+			if (view->document && view->document->focus_element)
+				html_document_update_focus_element (view->document, NULL);
 			view->root = new_box;
-		else {
+		 } else {
 			html_box_append_child (parent_box, new_box);
 
 			/* Mark the boxes that need to be relayouted */
@@ -687,8 +705,18 @@ html_view_button_release (GtkWidget *widget, GdkEventButton *event)
 static gboolean
 html_view_focus_in (GtkWidget *widget, GdkEventFocus *event)
 {
+	HtmlView *view;
+	DomElement *element;
+
 	GTK_WIDGET_SET_FLAGS (widget, GTK_HAS_FOCUS);
 
+	view = HTML_VIEW (widget);
+
+	element = html_view_get_saved_focus (view);
+	if (element) {
+		html_document_update_focus_element (view->document, element);
+		html_view_scroll_to_node (view, DOM_NODE (view->document->focus_element), HTML_VIEW_SCROLL_TO_BOTTOM);
+	}
 	return FALSE;
 }
 
@@ -702,6 +730,7 @@ html_view_focus_out (GtkWidget *widget, GdkEventFocus *event)
 	GTK_WIDGET_UNSET_FLAGS (widget, GTK_HAS_FOCUS);
 		
 	if (view->document->focus_element) {
+		html_view_set_saved_focus (view);
 		if (GTK_CONTAINER (widget)->focus_child == NULL)
 			html_document_update_focus_element (view->document, NULL);
 	}
@@ -718,11 +747,24 @@ html_view_key_press (GtkWidget *widget, GdkEventKey *event)
 }
 #endif
 
+static void
+focus_element_destroyed (gpointer data)
+{
+	g_object_set_data (G_OBJECT (data), "saved-focus", NULL);
+}
+
 static void
 html_view_finalize (GObject *object)
 {
 	HtmlView *view = HTML_VIEW (object);
+	gpointer saved_focus;
 
+	saved_focus = g_object_get_data (G_OBJECT (view), "saved-focus");
+	if (saved_focus) {
+		g_object_weak_unref (G_OBJECT (saved_focus),
+			   	     (GWeakNotify) focus_element_destroyed,
+				     view);
+	}
 	g_object_unref (view->document);
 
 	if (G_OBJECT_CLASS (parent_class)->finalize)
@@ -735,7 +777,7 @@ html_view_focus (GtkWidget *widget, GtkDirectionType direction)
 	HtmlView *view = HTML_VIEW (widget);
 	HtmlBox *box;
 	DomElement *old_focus_element = view->document->focus_element;
-	
+
 	if (view->document == NULL || view->document->dom_document == NULL) {
 		if (!GTK_WIDGET_HAS_FOCUS (widget)) {
 			gtk_widget_grab_focus (widget);
@@ -746,15 +788,9 @@ html_view_focus (GtkWidget *widget, GtkDirectionType direction)
 
 	if (direction == GTK_DIR_TAB_FORWARD) {
 		html_document_update_focus_element (view->document, html_focus_iterator_next_element (view->document->dom_document, view->document->focus_element));
-				
+			
 		if (view->document->focus_element) {
-			html_view_scroll_to_node (view, DOM_NODE (view->document->focus_element), HTML_VIEW_SCROLL_TO_BOTTOM);
-
-			box = html_view_find_layout_box (view, DOM_NODE (view->document->focus_element), FALSE);
-			if (box && HTML_IS_BOX_EMBEDDED (box) && HTML_BOX_EMBEDDED (box)) 
-				gtk_widget_child_focus (HTML_BOX_EMBEDDED (box)->widget, GTK_DIR_TAB_FORWARD);
-			else
-				gtk_widget_grab_focus (widget);
+			html_view_focus_element (view);
 			return TRUE;
 		}
 	}
@@ -762,13 +798,7 @@ html_view_focus (GtkWidget *widget, GtkDirectionType direction)
 		html_document_update_focus_element (view->document, html_focus_iterator_prev_element (view->document->dom_document, view->document->focus_element));
 
 		if (view->document->focus_element) {
-			html_view_scroll_to_node (view, DOM_NODE (view->document->focus_element), HTML_VIEW_SCROLL_TO_BOTTOM);
-
-			box = html_view_find_layout_box (view, DOM_NODE (view->document->focus_element), FALSE);
-			if (box && HTML_IS_BOX_EMBEDDED (box)) 
-				gtk_widget_child_focus (HTML_BOX_EMBEDDED (box)->widget, GTK_DIR_TAB_FORWARD);
-			else
-				gtk_widget_grab_focus (widget);
+			html_view_focus_element (view);
 			return TRUE;
 		}
 
@@ -836,6 +866,12 @@ html_view_real_move_cursor (HtmlView *html_view, GtkMovementStep step, gint coun
 	}
 }
 
+static void
+html_view_real_activate (HtmlView *view)
+{
+	html_event_activate (view);
+}
+
 static void
 html_view_add_move_binding (GtkBindingSet *binding_set, guint keyval, guint modmask, GtkMovementStep step, gint count)
 {
@@ -1014,6 +1050,7 @@ html_view_class_init (HtmlViewClass *klass)
 	container_class->set_focus_child = html_view_set_focus_child;
 	
 	klass->move_cursor = html_view_real_move_cursor;
+	klass->activate = html_view_real_activate;
 	
 	view_signals [MOVE_CURSOR] =
 		g_signal_new ("move_cursor",
@@ -1048,12 +1085,20 @@ html_view_class_init (HtmlViewClass *klass)
 			      1,
 			      G_TYPE_STRING);
 
+	view_signals [ACTIVATE] =
+		g_signal_new ("ACTIVATE",
+			      G_TYPE_FROM_CLASS (object_class),
+			      G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
+			      G_STRUCT_OFFSET (HtmlViewClass, activate),
+			      NULL, NULL,
+			      html_marshal_VOID__VOID,
+			      G_TYPE_NONE, 0);
+	widget_class->activate_signal = view_signals[ACTIVATE];
+
 	html_view_add_move_binding (binding_set, GDK_Page_Down, 0,
 				    GTK_MOVEMENT_PAGES, 1);
 	html_view_add_move_binding (binding_set, GDK_Page_Up, 0,
 				    GTK_MOVEMENT_PAGES, -1);
-	html_view_add_move_binding (binding_set, GDK_space, 0,
-				    GTK_MOVEMENT_PAGES, 1);
 	html_view_add_move_binding (binding_set, GDK_Home, 0,
 				    GTK_MOVEMENT_BUFFER_ENDS, -1);
 	html_view_add_move_binding (binding_set, GDK_End, 0,
@@ -1085,6 +1130,7 @@ html_view_init (HtmlView *view)
 	  
 	view->node_table = g_hash_table_new (g_direct_hash, g_direct_equal);
 	view->document = NULL;
+	view->root = NULL;
 	view->relayout_idle_id = 0;
 	view->relayout_timeout_id = 0;
 	view->magnification = 1.0;
@@ -1128,8 +1174,6 @@ html_view_new (void)
 {
 	HtmlView *view = HTML_VIEW (g_object_new (HTML_TYPE_VIEW, NULL));	
 
-	view->document = NULL;
-
 	/* This creates the adjustments needed */
  	gtk_layout_set_hadjustment (GTK_LAYOUT (view), NULL);
  	gtk_layout_set_vadjustment (GTK_LAYOUT (view), NULL);
@@ -1137,6 +1181,49 @@ html_view_new (void)
 	return GTK_WIDGET (view);
 }
 
+static void
+html_view_set_saved_focus (HtmlView *view)
+{
+	gpointer saved_focus;
+
+	saved_focus = g_object_get_data (G_OBJECT (view), "saved-focus");
+	if (saved_focus) {
+		g_object_weak_unref (G_OBJECT (saved_focus),
+				     (GWeakNotify) focus_element_destroyed,
+				     view);
+	}
+	g_object_weak_ref (G_OBJECT (view->document->focus_element),
+			   (GWeakNotify) focus_element_destroyed,
+			   view);
+	g_object_set_data (G_OBJECT (view), "saved-focus", view->document->focus_element);
+}
+
+static DomElement* 
+html_view_get_saved_focus (HtmlView *view)
+{
+	gpointer saved_focus;
+	
+	saved_focus = g_object_get_data (G_OBJECT (view), "saved-focus");
+	if (saved_focus)
+		return DOM_ELEMENT (saved_focus);
+	else
+		return NULL;
+}
+
+static void
+html_view_focus_element (HtmlView *view)
+{
+	HtmlBox *box;
+
+	html_view_scroll_to_node (view, DOM_NODE (view->document->focus_element), HTML_VIEW_SCROLL_TO_BOTTOM);
+
+	box = html_view_find_layout_box (view, DOM_NODE (view->document->focus_element), FALSE);
+	if (box && HTML_IS_BOX_EMBEDDED (box)) 
+		gtk_widget_child_focus (HTML_BOX_EMBEDDED (box)->widget, GTK_DIR_TAB_FORWARD);
+	else
+		gtk_widget_grab_focus (GTK_WIDGET (view));
+
+}
 
 #ifdef ENABLE_ACCESSIBILITY
 
@@ -1167,24 +1254,12 @@ html_view_get_accessible (GtkWidget *widget)
 			 * Specify what factory to use to create accessible
 			 * objects
 			 */
-			atk_registry_set_factory_type (registry, 
-						       HTML_TYPE_VIEW,
-						       HTML_TYPE_VIEW_ACCESSIBLE_FACTORY);
-			atk_registry_set_factory_type (registry,
-						       HTML_TYPE_BOX_BLOCK,
-						       HTML_TYPE_BOX_BLOCK_ACCESSIBLE_FACTORY);
-			atk_registry_set_factory_type (registry,
-						       HTML_TYPE_BOX_EMBEDDED,
-						       HTML_TYPE_BOX_EMBEDDED_ACCESSIBLE_FACTORY);
-			atk_registry_set_factory_type (registry,
-						       HTML_TYPE_BOX,
-						       HTML_TYPE_BOX_ACCESSIBLE_FACTORY);
-			atk_registry_set_factory_type (registry,
-						       HTML_TYPE_BOX_TABLE,
-						       HTML_TYPE_BOX_TABLE_ACCESSIBLE_FACTORY);
-			atk_registry_set_factory_type (registry,
-						       HTML_TYPE_BOX_TEXT,
-						       HTML_TYPE_BOX_TEXT_ACCESSIBLE_FACTORY);
+			HTML_ACCESSIBLE_SET_FACTORY (HTML_TYPE_VIEW, html_view_accessible);
+			HTML_ACCESSIBLE_SET_FACTORY (HTML_TYPE_BOX_BLOCK, html_box_block_accessible);
+			HTML_ACCESSIBLE_SET_FACTORY (HTML_TYPE_BOX_EMBEDDED, html_box_embedded_accessible);
+			HTML_ACCESSIBLE_SET_FACTORY (HTML_TYPE_BOX, html_box_accessible);
+			HTML_ACCESSIBLE_SET_FACTORY (HTML_TYPE_BOX_TABLE, html_box_table_accessible);
+			HTML_ACCESSIBLE_SET_FACTORY (HTML_TYPE_BOX_TEXT, html_box_text_accessible);
 		}
 		first_time = FALSE;
 	}
